
"""
Helpers for loading configurations and in-memory configuration state.

To make use of this module, you need `PyYAML` and `databind.json` installed.
"""

from ._loader import ConfigLoader

__all__ = ['ConfigLoader']




import logging
import types
import typing as t
import warnings

from nr.util.generic import T

logger = logging.getLogger(__name__)


class PluginRegistry:
  """ A helper class to register plugins associated with a unique identifier and later access them. """

  def __init__(self) -> None:
    self._groups: dict[t.Hashable, list[tuple[str, t.Any]]] = {}

  @t.overload
  def group(self, group: t.Hashable, base_class: type[T]) -> t.Iterable[tuple[str, T]]: ...

  @t.overload
  def group(self, group: t.Hashable) -> t.Iterable[tuple[str, t.Any]]: ...

  def group(self, group, base_class=None):
    plugins = list(self._groups.get(group, []))
    for idx, (plugin_name, plugin) in enumerate(plugins):
      if isinstance(plugin, (types.FunctionType, types.MethodType, types.LambdaType)):
        try:
          plugin = plugin()
        except:
          logger.exception('Unable to instanciate plugin "%s" (in group: "%s")', plugin_name, group)
          continue
      plugins[idx] = (plugin_name, plugin)

    if base_class is not None:
      assert isinstance(base_class, type), type(base_class)
      for idx, (plugin_name, plugin) in enumerate(plugins):
        if not isinstance(plugin, base_class):
          logger.error(
            'Plugin "%s" (registered in group: "%s") is not an instance of %s',
            plugin_name, group, base_class.__name__
          )
          continue
        plugins[idx] = (plugin_name, plugin)

    return plugins

  def register(self, group: t.Hashable, name: str, plugin: t.Any | t.Callable[[], t.Any]) -> None:
    """ Register a plugin under the specified *group* and *name*. If the *plugin* is callable (i.e. a function,
    method or lambda), it will be called with no arguments everytime #group() is called with the expectation to
    return an actual instance of the plugin. Registering multiple plugins with the same *name* in the same *group*
    causes that plugin to be registered multiple times (instead of replacing the previously registered plugin). """

    self._groups.setdefault(group, []).append((name, plugin))




import typing as t

from nr.util.generic import T_co
from ._supplier import Supplier


class Once(Supplier[T_co]):

  def __init__(self, supplier: Supplier[T_co]) -> None:
    self._supplier = supplier
    self._cached: bool = False
    self._value: T_co | None = None

  def __repr__(self) -> str:
    return f'Once({self._supplier!r})'

  def __call__(self) -> T_co:
    if not self._cached:
      self._value = self._supplier()
      self._cached = True
    return self._value

  def get(self, resupply: bool = False) -> T_co:
    if resupply:
      self._cached = False
    return self()



import dataclasses
import typing as t
import weakref
from collections.abc import Mapping

from nr.util.singleton import NotSet

K = t.TypeVar('K', bound=t.Hashable)
N = t.TypeVar('N')
E = t.TypeVar('E')


class DiGraph(t.Generic[K, N, E]):
  """
  Represents a directed graph.

  @generic N: The type of value stored for each node in the graph. All nodes in a graph are unique.
  @generic E: The type of value stored for each edge in the graph. Edge values may not be unique.
  """

  def __init__(self):
    """
    Create a new empty directed graph.
    """

    self._nodes: dict[K, '_Node[K, N]'] = {}
    self._roots: dict[K, None] = {}
    self._leafs: dict[K, None] = {}
    self._edges: dict[tuple[K, K], E] = {}
    self._nodesview = NodesView(self)
    self._edgesview = EdgesView(self)

  def add_node(self, node_id: K, value: N) -> None:
    """
    Add a node to the graph. Overwrites the existing node value if the *node_id* already exists in the graph,
    but keeps its edges intact.
    """

    existing_node = self._nodes.get(node_id, NotSet.Value)
    if existing_node is NotSet.Value:
      predecessors, successors = {}, {}
      self._roots[node_id] = None
      self._leafs[node_id] = None
    else:
      predecessors, successors = existing_node.predecessors, existing_node.successors
    self._nodes[node_id] = _Node(value, predecessors, successors)

  def add_edge(self, node_id1: K, node_id2: K, value: E) -> None:
    """
    Adds a directed edge from *node_id1* to *node_id2* to the graph, storing the given value along the edge.
    Overwrites the value if the edge already exists. The edge's nodes must be present in the graph.

    @raises UnknownNodeError: If one of the nodes don't exist in the graph.
    """

    node1, node2 = self._get_node(node_id1), self._get_node(node_id2)
    self._edges[(node_id1, node_id2)] = value
    node1.successors[node_id2] = None
    node2.predecessors[node_id1] = None
    self._leafs.pop(node_id1, None)
    self._roots.pop(node_id2, None)

  @property
  def nodes(self) -> 'NodesView[K, N]':
    """
    Returns a view on the nodes in the graph.
    """

    return self._nodesview

  @property
  def edges(self) -> 'EdgesView[K, E]':
    """
    Returns a view on the edges in the graph.
    """

    return self._edgesview

  @property
  def roots(self) -> t.KeysView[K]:
    """
    Return the nodes of the graph that have no predecessors.
    """

    return self._roots.keys()

  @property
  def leafs(self) -> t.KeysView[K]:
    """
    Return the nodes of the graph that have no successors.
    """

    return self._leafs.keys()

  def predecessors(self, node_id: K) -> t.KeysView[K]:
    """
    Returns a sequence of the given node's predecessor node IDs.

    @raises UnknownNodeError: If the node does not exist.
    """

    return self._get_node(node_id).predecessors.keys()

  def successors(self, node_id: K) -> t.KeysView[K]:
    """
    Returns a sequence of the given node's successor node IDs.

    @raises UnknownNodeError: If the node does not exist.
    """

    return self._get_node(node_id).successors.keys()

  # Internal

  def _get_node(self, node_id: K) -> '_Node[K, N]':
    try:
      return self._nodes[node_id]
    except KeyError:
      raise UnknownNodeError(node_id)


@dataclasses.dataclass
class _Node(t.Generic[K, N]):
  value: N
  predecessors: dict[K, None]
  successors: dict[K, None]


class NodesView(Mapping[K, N]):

  def __init__(self, g: DiGraph[K, N, t.Any]) -> None:
    self._g = weakref.ref(g)
    self._nodes = g._nodes

  def __repr__(self) -> str:
    return f'<NodesView count={len(self)}>'

  def __contains__(self, node_id: object) -> bool:
    return node_id in self._nodes

  def __len__(self) -> int:
    return len(self._nodes)

  def __iter__(self) -> t.Iterator[K]:
    return iter(self._nodes)

  def __getitem__(self, key: K) -> N:
    try:
      return self._nodes[key].value
    except KeyError:
      raise UnknownNodeError(key)

  def __setitem__(self, key: K, value: N) -> None:
    g = self._g()
    assert g is not None
    g.add_node(key, value)

  def __delitem__(self, key: K) -> None:
    g = self._g()
    assert g is not None
    node = g._nodes.pop(key)
    for pred in node.predecessors:
      g._nodes[pred].successors.pop(key)
      del g._edges[(pred, key)]
    for succ in node.successors:
      g._nodes[succ].predecessors.pop(key)
      del g._edges[(key, succ)]
    g._roots.pop(key, None)
    g._leafs.pop(key, None)


class EdgesView(Mapping[tuple[K, K], E]):

  def __init__(self, g: DiGraph[K, t.Any, E]) -> None:
    self._g = weakref.ref(g)
    self._edges = g._edges

  def __repr__(self) -> str:
    return f'<EdgesView count={len(self)}>'

  def __contains__(self, edge: object) -> bool:
    return edge in self._edges

  def __len__(self) -> int:
    return len(self._edges)

  def __iter__(self) -> t.Iterator[tuple[K, K]]:
    return iter(self._edges)

  def __getitem__(self, key: tuple[K, K]) -> E:
    try:
      return self._edges[key]
    except KeyError:
      raise UnknownEdgeError(key)

  def __setitem__(self, key: tuple[K, K], value: E) -> None:
    g = self._g()
    assert g is not None
    g.add_edge(key[0], key[1], value)

  def __delitem__(self, key: tuple[K, K]) -> None:
    del self._edges[key]


class UnknownNodeError(KeyError):
  pass


class UnknownEdgeError(KeyError):
  pass




from ._consumer import Consumer
from ._once import Once
from ._predicate import Predicate
from ._supplier import Supplier, ContextSupplier

__all__ = ['Consumer', 'Once', 'Predicate', 'Supplier', 'ContextSupplier']




import threading
import typing as t


class AtomicCounter:
  """
  A thread-safe counter.
  """

  def __init__(self) -> None:
    self._value = 0
    self._lock = threading.Lock()
    self._cond = threading.Condition(self._lock)

  def get(self) -> int:
    """
    Get the current value of the counter.
    """

    with self._cond:
      return self._value

  def inc(self) -> None:
    """
    Increment the counter by one.
    """

    with self._cond:
      self._value += 1

  def dec(self) -> None:
    """
    Decrement the counter by one. Raises a #ValueError if attempting to decrement a value if zero.
    """

    with self._cond:
      if self._value == 0:
        raise ValueError('cannot decrement AtomicCounter below 0')
      self._value -= 1
      self._cond.notify_all()

  def join(self, timeout: t.Optional[float] = None) -> None:
    """
    Block until the counter reaches zero, or until the timeout expires.
    """

    with self._cond:
      self._cond.wait_for(lambda: self._value == 0, timeout)




import typing as t
from collections.abc import Sequence
from nr.util.generic import T


def longest_common_substring(
  seq1: Sequence[T],
  seq2: Sequence[T],
  *args: Sequence[T],
  start_only: bool = False,
) -> Sequence[T]:
  """ Finds the longest common contiguous sequence of elements in *seq1* and *seq2* and returns it. """

  longest: tuple[int, int] = (0, 0)
  for i in (0,) if start_only else range(len(seq1)):
    for j in (0,) if start_only else range(len(seq2)):
      k = 0
      while (i + k < len(seq1) and (j + k) < len(seq2)) and seq1[i + k] == seq2[j + k]:
        k += 1
      if k > longest[1] - longest[0]:
        longest = (i, i + k)

  result = seq1[longest[0]:longest[1]]
  if args:
    result = longest_common_substring(result, *args, start_only=start_only)
  return result




from ._scanner import Cursor, Scanner, Seek
from ._tokenizer.extractor import TokenExtractor
from ._tokenizer.ruleset import RuleSet
from ._tokenizer.tokenizer import ProxyToken, Token, Tokenizer, TokenizationError, UnexpectedTokenError
from . import rules


__all__ = ['Cursor', 'Scanner', 'Seek']




""" Utilities for filesystem operations. """

from ._atomic import atomic_swap, atomic_write
from ._discovery import get_file_in_directory

__all__ = [
  'atomic_swap',
  'atomic_write',
  'get_file_in_directory',
]




""" Helpers to implement a plugin infrastructure in Python. """

import logging
import pkg_resources
import typing as t

from nr.util.generic import T

logger = logging.getLogger(__name__)


def load_plugins_from_entrypoints(
  group: str,
  base_class: type[T],
  constructor: t.Callable[[type[T]], T] | None = None,
  do_raise: bool = False,
  names: list[str] | None = None,
  filter: t.Callable[[pkg_resources.EntryPoint], bool] | None = None,
) -> dict[str, T]:
  """ Loads plugins from an entrypoint group. All entrypoints must point to a class of the specified *base_class*.
  If *do_raise* is `True` and an entrypoint does not point to a subclass of the given type, a #RuntimeError is
  raised, otherwise a warning is printed. If the entrypoint cannot be imported and *do_raise* is `False`, a warning
  will be printed as well.

  The *constructor* will be used to create instances of the *base_class*. If it is omitted, the instances will be
  created without arguments. If the construction fails and *do_raise* is `False`, a warning will be printed.

  By passing a predicate for the *filter* argument, you can filter which entry points in the group are considered
  for loading in the first place. """

  if not isinstance(base_class, type):
    raise TypeError(f'base_class must be a type, got {type(base_class).__name__}')

  result = {}
  for ep in pkg_resources.iter_entry_points(group):
    if names is not None and ep.name not in names:
      continue
    if filter is not None and not filter(ep):
      continue

    try:
      cls = ep.load()
    except ImportError:
      if do_raise:
        raise
      logger.exception(f'Unable to load entrypoint "%s" due to ImportError', ep)
      continue

    if not isinstance(cls, type) or not issubclass(cls, base_class):
      message = f'Entrypoint "%s" does not point to a %s subclass (got value: %r)'
      if do_raise:
        raise RuntimeError(message % (ep, base_class.__name__, cls))
      logger.error(message, ep, base_class.__name__, cls)
      continue

    try:
      result[ep.name] = constructor(cls) if constructor else cls()
    except Exception:
      message = f'Instance from entrypoint "%s" could not be created'
      if do_raise:
        raise RuntimeError(message % (ep,))
      logger.exception(message, ep)

  if names is not None and (missing_plugins := set(names) - result.keys()):
    msg = f'missing plugin{"" if len(missing_plugins) == 1 else "s"} in entrypoint group "{group}": {missing_plugins}'
    raise RuntimeError(msg)

  return result




import os
import sys

if __name__ == '__main__':
  # Ensure that the parent directory is not in sys.path.
  norm = lambda x: os.path.normpath(os.path.abspath(x))
  dirname = os.path.dirname(norm(__file__))
  sys.path[:] = [x for x in sys.path if norm(x) != dirname]
  del norm, dirname


import ctypes
import io
import json
import re
import shutil
import shlex
import subprocess
import tempfile
import traceback
import typing as t


if os.name == 'nt':
  import ctypes.wintypes as wintypes
  windll = ctypes.windll  # type: ignore
  WinError = ctypes.WinError  # type: ignore
  get_last_error = ctypes.get_last_error  # type: ignore
  class winapi:
    _WaitForSingleObject = windll.kernel32.WaitForSingleObject
    _WaitForSingleObject.restype = wintypes.DWORD
    _WaitForSingleObject.argtypes = [wintypes.HANDLE, wintypes.DWORD]

    @staticmethod
    def WaitForSingleObject(handle, ms=0):
      return winapi._WaitForSingleObject(handle, ms)

    _GetExitCodeProcess = windll.kernel32.GetExitCodeProcess
    _GetExitCodeProcess.restype = wintypes.BOOL
    _GetExitCodeProcess.argtypes = [wintypes.HANDLE, ctypes.POINTER(wintypes.DWORD)]

    @staticmethod
    def GetExitCodeProcess(handle):
      result = wintypes.DWORD()
      success = winapi._GetExitCodeProcess(handle, ctypes.byref(result))
      if not success:
        raise WinError(get_last_error())
      return result.value

    _MessageBox = windll.user32.MessageBoxW
    _MessageBox.restype = ctypes.c_int
    _MessageBox.argtypes = [wintypes.HWND, wintypes.LPWSTR, wintypes.LPWSTR, wintypes.UINT]

    @staticmethod
    def MessageBox(hwnd, text, caption, type):
      return winapi._MessageBox(hwnd, text, caption, type)

    class _SHELLEXECUTEINFO(ctypes.Structure):
      _fields_ = [
        ('cbSize', wintypes.DWORD),
        ('fMask', wintypes.ULONG),
        ('hwnd', wintypes.HWND),
        ('lpVerb', wintypes.LPCSTR),
        ('lpFile', wintypes.LPCSTR),
        ('lpParameters', wintypes.LPCSTR),
        ('lpDirectory', wintypes.LPCSTR),
        ('nShow', ctypes.c_int),
        ('hInstApp', wintypes.HINSTANCE),
        ('lpIDList', wintypes.LPVOID),
        ('lpClass', wintypes.LPCSTR),
        ('hkeyClass', wintypes.HKEY),
        ('dwHotKey', wintypes.DWORD),
        ('DUMMYUNIONNAME', wintypes.HANDLE),
        ('hProcess', wintypes.HANDLE),
      ]

    _ShellExecuteEx = windll.shell32.ShellExecuteEx
    _ShellExecuteEx.restype = wintypes.BOOL
    _ShellExecuteEx.argtypes = [ctypes.POINTER(_SHELLEXECUTEINFO)]

    SW_HIDE = 0
    SW_MAXIMIMIZE = 3
    SW_MINIMIZE = 6
    SW_RESTORE = 9
    SW_SHOW = 5
    SW_SHOWDEFAULT = 10
    SW_SHOWMAXIMIZED = 3
    SW_SHOWMINIMIZED = 2
    SW_SHOWMINNOACTIVE = 7
    SW_SHOWNA = 8
    SW_SHOWNOACTIVE = 4
    SW_SHOWNORMAL = 1

    @staticmethod
    def ShellExecuteEx(hwnd=None, verb='', file='', parameters=None,
                       directory=None, show=SW_SHOW, mask=0):  # TODO: More parameters
      data = winapi._SHELLEXECUTEINFO()
      data.cbSize = ctypes.sizeof(data)
      data.fMask = mask
      data.hwnd = hwnd
      data.lpVerb = verb.encode()
      data.lpFile = file.encode()
      data.lpParameters = parameters.encode()
      data.lpDirectory = directory.encode()
      data.nShow = show
      data.hInstApp = None
      data.lpIDList = None
      data.lpClass = None
      data.hkeyClass = None
      data.dwHotKey = 0
      data.DUMMYUNIONNAME = None
      data.hProcess = None
      result = winapi._ShellExecuteEx(ctypes.byref(data))
      if not result:
        raise WinError(get_last_error())
      return {'hInstApp': data.hInstApp, 'hProcess': data.hProcess}


def alert(*msg: str) -> None:
  # TODO (@NiklasRosenstein): Support GUI alerts for other systems.
  message = ' '.join(map(str, msg))
  print(message, file=sys.stderr)
  sys.stderr.flush()
  if os.name == 'nt':
    winapi.MessageBox(None, message, "Python", 0)


def quote(s: str) -> str:
  if os.name == 'nt' and os.sep == '\\':
    s = s.replace('"', '\\"')
    if re.search('\s', s) or any(c in s for c in '<>'):
      s = '"' + s + '"'
  else:
    s = shlex.quote(s)
  return s


def is_root() -> bool:
  if os.name == 'nt':
    try:
      return bool(windll.shell32.IsUserAnAdmin())
    except:
      traceback.print_exc()
      print("ctypes.windll.shell32.IsUserAnAdmin() failed -- "
            "assuming not an admin.", file=sys.stderr)
      sys.stderr.flush()
      return False
  elif os.name == 'posix':
    return os.getuid() == 0
  else:
    raise RuntimeError('Unsupported os: {!r}'.format(os.name))


def elevate(command: str | list[str], cwd: str | None = None, environ: t.Mapping[str, str] | None  = None) -> None:
  """
  Runs a command as an admin in the specified *cwd* and *environ*. On Windows, this creates a temporary directory where
  this information is stored temporarily so that the new process can launch the proper subprocess.
  """

  if isinstance(command, str):
    command = shlex.split(command)

  if os.name == 'nt':
    return _elevate_windows(command, cwd, environ)

  elif os.name == 'posix':
    command = ['sudo', '-E'] + list(command)
    sys.exit(subprocess.call(command))

  else:
    raise RuntimeError('Unsupported os: {!r}'.format(os.name))


def _elevate_windows(command, cwd, environ):
  assert os.name == 'nt'
  datadir = tempfile.mkdtemp()
  try:
    # TODO: Maybe we could also use named pipes and transfer them
    #       via the processdata.json to the elevated process.
    # This file will receive all the process information.
    datafile = os.path.join(datadir, 'processdata.json')
    data = {
      'command': command,
      'cwd': cwd or os.getcwd(),
      'environ': environ or os.environ.copy(),
      'outfile': os.path.join(datadir, 'out.bin')
    }
    with open(datafile, 'w') as fp:
      json.dump(data, fp)

    # Ensure the output file exists.
    open(data['outfile'], 'w').close()

    # Create the windows elevated process that calls this file. This
    # file will then know what to do with the information from the
    # process data directory.
    hProc = winapi.ShellExecuteEx(
      file=sys.executable,
      verb='runas',
      parameters=' '.join(map(quote, [os.path.abspath(__file__), '--windows-process-data', datadir])),
      directory=datadir,
      mask=64,
      show=winapi.SW_HIDE
      )['hProcess']

    # Read the output from the process and write it to our stdout.
    with open(data['outfile'], 'rb+', 0) as outfile:
      while True:
        hr = winapi.WaitForSingleObject(hProc, 40)
        while True:
          line = outfile.readline()
          if not line: break
          sys.stdout.buffer.write(line)
        if hr != 0x102: break

    return winapi.GetExitCodeProcess(hProc)

  finally:
    try:
      shutil.rmtree(datadir)
    except:
      print("ERROR: Unable to remove data directory of elevated process.")
      print("ERROR: Directory at \"{}\"".format(datadir))
      traceback.print_exc()


def _elevate_windows_elevated(datadir):
  assert os.name == 'nt'
  datafile = os.path.join(datadir, 'processdata.json')
  with open(datafile, 'r') as pdata_fp:
    data = json.load(pdata_fp)

  try:
    with open(data['outfile'], 'wb', 0) as fp:
      sys.stderr = sys.stdout = io.TextIOWrapper(fp)
      os.environ.update(data['environ'])
      return subprocess.call(data['command'], cwd=data['cwd'], stdout=fp, stderr=fp)
  except:
    alert(traceback.format_exc())
    sys.exit(1)


def main(argv=None, prog=None):
  import argparse
  parser = argparse.ArgumentParser(prog=prog)
  parser.add_argument('--windows-process-data',
    help='The path to a Windows process data directory. This is used to '
      'provide data for the elevated process since no environment variables '
      'can be via ShellExecuteEx().')
  args, unknown = parser.parse_known_args(argv)

  if args.windows_process_data:
    if not is_root():
      alert("--windows-process-data can only be used in an elevated process.")
      sys.exit(1)
    sys.exit(_elevate_windows_elevated(args.windows_process_data))
  elif unknown:
    elevate(unknown)
    sys.exit()
  else:
    parser.print_usage()


_entry_point = lambda: sys.exit(main())

if __name__ == '__main__':
  _entry_point()




from pathlib import Path


def get_file_in_directory(
  directory: Path,
  prefix: str,
  preferred: list[str],
  case_sensitive: bool = True,
) -> Path | None:
  """ Returns a file in *directory* that is either in the *preferred* list or starts with specified *prefix*. """

  if not case_sensitive:
    preferred = [x.lower() for x in preferred]

  choices = []
  for path in sorted(directory.iterdir()):
    if (case_sensitive and path.name in preferred) or (not case_sensitive and path.name.lower() in preferred):
      return path
    if path.name.startswith(prefix):
      choices.append(path)
  else:
    if choices:
      return choices[0]

  return None



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['util',
 'util.algorithm',
 'util.annotations',
 'util.atomic',
 'util.awsgi',
 'util.awsgi.launcher',
 'util.config',
 'util.digraph',
 'util.digraph.algorithm',
 'util.fs',
 'util.functional',
 'util.generic',
 'util.git',
 'util.io',
 'util.keyvalue',
 'util.parsing',
 'util.parsing._tokenizer',
 'util.plugins',
 'util.process',
 'util.proxy',
 'util.safearg',
 'util.singleton',
 'util.task',
 'util.weak']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'nr.util',
    'version': '0.4.8',
    'description': 'General purpose Python utility library.',
    'long_description': '# nr.util\n\nGeneral purpose Python utility library.\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)





import typing as t
import weakref

from nr.util.generic import T


class WeakProperty(t.Generic[T]):

  def __init__(self, attr_name: str, once: bool = False) -> None:
    self._name = attr_name
    self._once = once
    self._value: t.Optional[weakref.ReferenceType[T]] = None

  def __set__(self, instance: t.Any, value: T) -> None:
    has_value: t.Optional[weakref.ReferenceType[T]] = getattr(instance, self._name, None)
    if self._once and has_value is not None:
      raise RuntimeError('property can not be set more than once')
    setattr(instance, self._name, weakref.ref(value))

  def __get__(self, instance: t.Any, owner: t.Any) -> T:
    if instance is None:
      raise AttributeError()
    has_value: t.Optional[weakref.ReferenceType[T]] = getattr(instance, self._name, None)
    if has_value is None:
      raise AttributeError('property value is not set')
    value = has_value()
    if value is None:
      raise RuntimeError('lost weak reference')
    return value


class OptionalWeakProperty(WeakProperty[t.Optional[T]]):

  def __set__(self, instance: t.Any, value: t.Optional[T]) -> None:
    has_value: t.Optional[weakref.ReferenceType[T]] = getattr(instance, self._name, None)
    if self._once and has_value is not None:
      raise RuntimeError('property can not be set more than once')
    setattr(instance, self._name, weakref.ref(value) if value is not None else None)

  def __get__(self, instance: t.Any, owner: t.Any) -> t.Optional[T]:
    if instance is None:
      raise AttributeError()
    try:
      return super().__get__(instance, owner)
    except AttributeError:
      return None


if t.TYPE_CHECKING:
  @t.overload
  def weak_property(attr_name: str, once: bool = False, optional: t.Literal[False] = False) -> T: ...
  @t.overload
  def weak_property(attr_name: str, once: bool = False, optional: t.Literal[True] = True) -> T | None: ...


def weak_property(attr_name: str, once: bool = False, optional: bool = False) -> T | None:
  return t.cast(T, WeakProperty(attr_name, once) if optional else OptionalWeakProperty(attr_name, once))




""" A selection of useful algorithms. """

# TODO For backwards compatibility, remove in major version.
from .longest_common_substring import longest_common_substring

__all__ = ['longest_common_substring']




import dataclasses
import sys
import typing as t

from . import AWSGILauncher


@dataclasses.dataclass
class UvicornLauncher(AWSGILauncher):
  """
  Launches your ASGI/WSGI application via Uvicorn.
  """

  # TODO (@NiklasRosenstein): Ensure that Uvicorn access/error logs end up in var/log

  #: Bind socket to this host.
  host: str | None = None

  #: Bind socket to this port.
  port: int | None = 8000

  #: Bind to a UNIX domain socket.
  unix_socket: str | None = None

  #: Number of worker processes.
  workers: int | None = None

  #: Enable auto-reload.
  reload: bool = False

  #: Event loop implementation.
  loop: str = 'auto'

  #: HTTP protocol implementation.
  http: str = 'auto'

  #: Additional keyword arguments for the Uvicorn invokation.
  kwargs: dict[str, t.Any] = dataclasses.field(default_factory=dict)

  #: The entrypoint to launch.
  entrypoint: str | None = None

  def launch(self, entrypoint: str, factory: bool = False) -> None:
    import uvicorn  # type: ignore
    try:
      sys.exit(uvicorn.run(
        entrypoint,
        host=self.host,
        port=self.port,
        uds=self.unix_socket,
        workers=self.workers,
        reload=self.reload,
        loop=self.loop,
        http=self.http,
        factory=factory,
      ))
    except KeyboardInterrupt:
      sys.exit(1)
